/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  GMime
 *  Copyright (C) 2000-2009 Jeffrey Stedfast
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1
 *  of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
 *  02110-1301, USA.
 */

#if defined _MSC_VER
#pragma warning(disable: 4996)
#endif

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <sys/types.h>

#include <gmime/gmime-part-charset.h>
#include <gmime/gmime-events.h>
#include <gmime/gmime-object.h>

#include <gmime/gmime-stream-null.h>
#include <gmime/gmime-stream-filter.h>
#include <gmime/gmime-filter-basic.h>
#include <gmime/gmime-filter-best.h>
#include <gmime/gmime-filter-charset.h>

#define d(x)


/**
 * SECTION: gmime-part-charset
 * @title: GMimePartCharset
 * @short_description: MIME parts with charset processing
 * @see_also:
 *
 * A #GMimePartCharset represents MIME leaf part (meaning it has no
 * sub-parts) with charset. It handle charset conversion, mainly from/to UNICODE charsets
 **/

/* GObject class methods */
static void g_mime_part_charset_class_init (GMimePartCharsetClass *klass);
static void g_mime_part_charset_init (GMimePartCharset *mime_part, GMimePartCharsetClass *klass);
static void g_mime_part_charset_finalize (GObject *object);

/* GMimeObject class methods */
static void mime_part_charset_set_content_type (GMimeObject *object, GMimeContentType *content_type);

/* GMimePart class methods */
static void mime_part_charset_set_content_object (GMimePart *mime_part, GMimeDataWrapper *content);
static ssize_t write_headers_to_stream(GMimePart *mime_part, GMimeStream *stream);
static ssize_t write_contents_to_stream(GMimePart *mime_part, GMimeStream *stream);


/* GMimePartCharset class methods */
static void set_content_object_with_charset (GMimePartCharset *mime_part, GMimeDataWrapper *content, const char *content_charset);

static void content_type_changed (GMimeContentType *content_type, gpointer args, GMimeObject *object);


static GMimeObjectClass *parent_class = NULL;


GType
g_mime_part_charset_get_type (void)
{
	static GType type = 0;
	
	if (!type) {
		static const GTypeInfo info = {
			sizeof (GMimePartCharsetClass),
			NULL, /* base_class_init */
			NULL, /* base_class_finalize */
			(GClassInitFunc) g_mime_part_charset_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (GMimePartCharset),
			0,    /* n_preallocs */
			(GInstanceInitFunc) g_mime_part_charset_init,
		};
		
		type = g_type_register_static (GMIME_TYPE_PART, "GMimePartCharset", &info, 0);
	}
	
	return type;
}


static void
g_mime_part_charset_class_init (GMimePartCharsetClass *klass)
{
    GMimePartClass *part_class = GMIME_PART_CLASS(klass);
    GMimeObjectClass *object_class = GMIME_OBJECT_CLASS (klass);
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
	
	parent_class = g_type_class_ref (GMIME_TYPE_PART);
	
	gobject_class->finalize = g_mime_part_charset_finalize;
	
    object_class->set_content_type = mime_part_charset_set_content_type;

    part_class->set_content_object = mime_part_charset_set_content_object;
    part_class->write_headers_to_stream = write_headers_to_stream;
    part_class->write_contents_to_stream = write_contents_to_stream;

    klass->set_content_object_with_charset = set_content_object_with_charset;
}

static void
g_mime_part_charset_init (GMimePartCharset *mime_part, GMimePartCharsetClass *klass)
{
    mime_part->charset = NULL;
    mime_part->content_charset = NULL;
}

static void
g_mime_part_charset_finalize (GObject *object)
{
	GMimePartCharset *mime_part = (GMimePartCharset *) object;
	
	g_free (mime_part->charset);
    g_free (mime_part->content_charset);
	
	G_OBJECT_CLASS (parent_class)->finalize(object);
}

static const char *nontransportable_unicode_charsets[] = {
    "UCS-2", 
    "UCS-2BE", 
    "UCS-2LE",
    "UCS-4", 
    "UCS-4BE", 
    "UCS-4LE"
    "UTF-16BE", 
    "UTF-16LE"
    "UTF-32BE", 
    "UTF-32LE",
    "UCS-2-INTERNAL", 
    "UCS-4-INTERNAL"
};


static gboolean
is_nontransportable_unicode_charset(const char *charset)
{
    int i;
    gboolean result = FALSE;
    const char *iconv_charset;
    g_return_val_if_fail (charset, result);

    iconv_charset = g_mime_charset_iconv_name(charset);
    for (i = 0; i < G_N_ELEMENTS (nontransportable_unicode_charsets); i++) {
        if (0 == g_ascii_strcasecmp (nontransportable_unicode_charsets[i], iconv_charset)) {
            result = TRUE;
            break;
        }
    }

    return result;
}

static void
mime_part_charset_set_content_type (GMimeObject *object, GMimeContentType *content_type)
{
    if (object->content_type) {
        g_mime_event_remove (object->content_type->priv, (GMimeEventCallback) content_type_changed, object);
    }

    GMIME_OBJECT_CLASS(parent_class)->set_content_type(object, content_type);

    if (object->content_type) {
        g_mime_event_add (content_type->priv, (GMimeEventCallback) content_type_changed, object);
    }
}

static void
content_type_changed (GMimeContentType *content_type, gpointer args, GMimeObject *object)
{
    GMimePartCharset *me = (GMimePartCharset *)object;
    const char *charset;
    gboolean changed;

    charset = g_mime_content_type_get_parameter(content_type, "charset");
    if (charset && me->charset) {
        changed = 0 != g_ascii_strcasecmp(charset, me->charset);
    } else {
        changed = charset != me->charset;
    }

    if (changed && charset) {
        g_free(me->charset);
        me->charset = g_strdup(charset);
    }
}

static void
mime_part_charset_set_content_object(GMimePart *mime_part, GMimeDataWrapper *content)
{
    GMimePartCharset *me = (GMimePartCharset *)mime_part;
    GMIME_PART_CLASS(parent_class)->set_content_object(mime_part, content);
    g_free(me->content_charset);
}

static void 
set_content_object_with_charset(GMimePartCharset *mime_part, GMimeDataWrapper *content, const char *content_charset)
{
    GMimePart *mp = (GMimePart *)mime_part;
    g_mime_part_set_content_object(mp, content);
    g_free(mime_part->content_charset);
    if (mp->content && content_charset) {
        mime_part->content_charset = g_strdup(content_charset);
    }
}


static ssize_t
write_content (GMimePartCharset *part, GMimeStream *stream)
{
    ssize_t nwritten, total = 0;
    GMimeStream *filtered_stream;
    const char *filename;
    GMimePart *mime_part = (GMimePart *)part;

    if (!mime_part->content)
        return 0;

    filtered_stream = g_mime_stream_filter_new (stream);

    if (part->charset && part->content_charset && 0 != g_ascii_strcasecmp(part->charset, part->content_charset)) {
        char *targetCharset = g_strconcat(part->charset, "//TRANSLIT", NULL);
        GMimeFilter *charset_filter = g_mime_filter_charset_new (part->content_charset, targetCharset);
        g_free(targetCharset);

        /* P.Dostal - if filter is not found (e.g. encoding is unknown for iconv) and conversion is from unicode, convert to utf-8 */
        if (NULL == charset_filter && FALSE != is_nontransportable_unicode_charset(part->content_charset)) {
            charset_filter = g_mime_filter_charset_new(part->content_charset, "utf-8//TRANSLIT");
        }
                
        /* P.Dostal - if filter is not found (e.g. encoding is unknown for iconv), do not try to use it */
        if (NULL != charset_filter) {
            g_mime_stream_filter_add (GMIME_STREAM_FILTER (filtered_stream), charset_filter);
            g_object_unref(charset_filter);
        }
    }

    switch (mime_part->encoding) {
        case GMIME_CONTENT_ENCODING_UUENCODE:
            filename = g_mime_part_get_filename (mime_part);
            nwritten = g_mime_stream_printf (stream, "begin 0644 %s\n",
                filename ? filename : "unknown");
            if (nwritten == -1)
                return -1;

            total += nwritten;

            /* fall thru... */
        case GMIME_CONTENT_ENCODING_QUOTEDPRINTABLE:
        case GMIME_CONTENT_ENCODING_BASE64:
            {
                GMimeFilter *filter = g_mime_filter_basic_new (mime_part->encoding, TRUE);
                g_mime_stream_filter_add (GMIME_STREAM_FILTER (filtered_stream), filter);
                g_object_unref (filter);
            }
            break;
        default:
            break;
    }

    nwritten = g_mime_data_wrapper_write_to_stream (mime_part->content, filtered_stream);
    g_mime_stream_flush (filtered_stream);
    g_object_unref (filtered_stream);

    if (nwritten == -1)
        return -1;

    total += nwritten;

    if (mime_part->encoding == GMIME_CONTENT_ENCODING_UUENCODE) {
        /* FIXME: get rid of this special-case x-uuencode crap */
        nwritten = g_mime_stream_write (stream, "end\n", 4);
        if (nwritten == -1)
            return -1;

        total += nwritten;
    }

    return total;
}


/**
 * g_mime_part_charset_new:
 *
 * Creates a new MIME Part object with a default content-type of
 * text/plain.
 *
 * Returns: an empty MIME Part object with a default content-type of
 * text/plain.
 **/
GMimePartCharset *
g_mime_part_charset_new (void)
{
	GMimePartCharset *mime_part;
	
	mime_part = g_object_newv (GMIME_TYPE_PART_CHARSET, 0, NULL);
	
	return mime_part;
}


/**
 * g_mime_part_charset_new_with_type:
 * @type: content-type string
 * @subtype: content-subtype string
 *
 * Creates a new MIME Part with a sepcified type.
 *
 * Returns: an empty MIME Part object with the specified content-type.
 **/
GMimePartCharset *
g_mime_part_charset_new_with_type (const char *type, const char *subtype)
{
	GMimeContentType *content_type;
	GMimePartCharset *mime_part;
	
	mime_part = g_object_newv (GMIME_TYPE_PART_CHARSET, 0, NULL);
	
	content_type = g_mime_content_type_new (type, subtype);
	g_mime_object_set_content_type (GMIME_OBJECT (mime_part), content_type);
	g_object_unref (content_type);
	
	return mime_part;
}


const char *g_mime_part_charset_get_content_charset(GMimePartCharset *mime_part)
{
    g_return_val_if_fail (GMIME_IS_PART_CHARSET (mime_part), NULL);

    return mime_part->content_charset;
}

void g_mime_part_charset_set_content_object_with_charset(GMimePartCharset *mime_part, GMimeDataWrapper *content, const char *content_charset)
{
    g_return_if_fail (GMIME_IS_PART_CHARSET (mime_part));
    GMIME_PART_CHARSET_GET_CLASS(mime_part)->set_content_object_with_charset(mime_part, content, content_charset);
}

const char *g_mime_part_charset_get_best_charset(GMimePartCharset *mime_part)
{
    GMimeStream *null, *filter_stream;
    GMimeFilter *best_filter;
    GMimeDataWrapper *content;
    const char *charset;
    g_return_val_if_fail (GMIME_IS_PART_CHARSET (mime_part), NULL);

    content = g_mime_part_get_content_object(GMIME_PART(mime_part));
    g_return_val_if_fail (content, NULL);

    null = g_mime_stream_null_new();
    filter_stream = g_mime_stream_filter_new(null);
    g_object_unref(null);
    best_filter = g_mime_filter_best_new(GMIME_FILTER_BEST_CHARSET);
    g_mime_data_wrapper_write_to_stream(content, filter_stream);
    g_object_unref(filter_stream);
    charset = g_mime_filter_best_charset((GMimeFilterBest *)best_filter);
    g_object_unref(best_filter);

    return charset;
}


ssize_t 
write_headers_to_stream(GMimePart *part, GMimeStream *stream)
{
    GMimeObject *object = (GMimeObject *)part;
    GMimePartCharset *mime_part = (GMimePartCharset *) part;
    ssize_t nwritten;

    if (mime_part->content_charset && !mime_part->charset) {
        if (is_nontransportable_unicode_charset(mime_part->content_charset)) {
            mime_part->charset = g_strdup(g_mime_charset_iconv_name("utf-8"));
        } else {
            mime_part->charset = g_strdup(mime_part->content_charset);
        }
        g_mime_object_set_content_type_parameter(object, "charset", mime_part->charset);
    }

    /* write the content headers */
    nwritten = g_mime_header_list_write_to_stream (object->headers, stream, mime_part->charset);

    return nwritten;
}

ssize_t 
write_contents_to_stream(GMimePart *part, GMimeStream *stream)
{
    GMimePartCharset *mime_part = (GMimePartCharset *)part;

    ssize_t nwritten = write_content(mime_part, stream);

    return nwritten;
}
